import 'package:flutter/material.dart';
import 'package:rapidfire/store/modern/dashboard/piechart.dart';
import '../../../Config.dart';
import 'colors.dart';
import 'size.dart';
import 'strings.dart';
import '../../../Db.dart';

class DashBoard extends StatefulWidget {
  @override
  _DashBoardState createState() => _DashBoardState();
}

class _DashBoardState extends State<DashBoard> {
  List<appContent> items = [];

  @override
  void didChangeDependencies() {
    if (items.isEmpty) addAppItems();
  }

  @override
  Widget build(BuildContext context) {
    var height = SizeConfig.getHeight(context);
    var width = SizeConfig.getWidth(context);
    double fontSize(double size) {
      return size * width / 414;
    }

    return Column(
      children: [
        Container(
          height: 300,
          child: Row(
            children: <Widget>[
              Expanded(
                flex: 5,
                child: Center(
                  child: SingleChildScrollView(
                    child: Column(
                      mainAxisAlignment: MainAxisAlignment.center,
                      children: items.map((data) {
                        return Container(
                          padding: EdgeInsets.symmetric(horizontal: width / 20, vertical: height / 90),
                          child: Row(
                            children: <Widget>[
                              Container(
                                margin: EdgeInsets.only(right: 10),
                                width: 12,
                                height: 12,
                                decoration: BoxDecoration(
                                    color: AppColors.pieColors[items.indexOf(data)], shape: BoxShape.circle),
                              ),
                              Row(
                                children: [
                                  Text(
                                    data.Title,
                                    style: TextStyle(
                                      fontSize: fontSize(16),
                                    ),
                                  ),
                                  SizedBox(
                                    width: 6,
                                  ),
                                  Text(
                                    "(" + data.DataCount.toString() + ")",
                                    style: TextStyle(
                                      color: Colors.grey,
                                      fontSize: fontSize(12),
                                    ),
                                  ),
                                ],
                              )
                            ],
                          ),
                        );
                      }).toList(),
                    ),
                  ),
                ),
              ),
              Expanded(
                flex: 6,
                child: Padding(
                  padding: EdgeInsets.only(right: width / 30),
                  child: Container(child: PieChart(items)),
                ),
              ),
            ],
          ),
        ),
      ],
    );
  }

  addAppItems() async {
    Models models = Models();
    for (var item in models.get().entries) {
      var modelRF = item.value as ModelRF;
      if (modelRF.syncable) {
        Db db = Db();
        int count = await db.count('SELECT * FROM ${modelRF.model.runtimeType.toString()}');
        items.add(appContent(Title: modelRF.displayName, DataCount: count));
      }
    }
    setState(() {});
  }
}

class appContent {
  String Title;
  int DataCount;
  appContent({this.Title, this.DataCount = 0});
}
