import 'package:flutter/material.dart';

//import 'package:searchable_dropdown/searchable_dropdown.dart';
import 'package:rapidfire/uiux/searchable_dropdown.dart';

import '../Dynamic.dart';
import '../GlobalScopeRF.dart';
import '../ValidationRF.dart';
import '../Extensions.dart';
import 'WidgetsRF.dart';
import 'StyleRF.dart';

enum DropdownType { simple, searchable }
enum DropdownStyle { normal, fluid }

class DropdownRF extends StatefulWidget {
  const DropdownRF({
    this.items,
    this.model,
    this.field,
    this.itemsOLD,
    this.value,
    this.bind,
    this.dropdownType = DropdownType.simple,
    this.maxLength,
    this.hintText,
    this.hint,
    this.caption,
    this.onSaved,
    this.validator,
    this.onFieldSubmitted,
    this.onChange,
    this.backColor = const Color(0x77f7f7f7),
    this.style = DropdownStyle.normal,
    this.textColor,
    this.textSize,
  });

  final items;
  final field;
  final model;
  final List<DropdownMenuItem> itemsOLD;
  final String value;
  final dynamic bind;
  final DropdownType dropdownType;
  final int maxLength;
  final String hintText;
  final String hint;
  final String caption;
  final FormFieldSetter<String> onSaved;
  final FormFieldValidator<String> validator;
  final ValueChanged<String> onFieldSubmitted;
  final Function onChange;
  final backColor;
  final style;
  final textColor;
  final textSize;

  @override
  _DropdownListRF createState() => _DropdownListRF();
}

class _DropdownListRF extends State<DropdownRF> {
  int _selectedIndex = 0;
  String _selectedValue = "";

  @override
  Widget build(BuildContext context) {
    var model = PageRF.of(context)?.model;

    if (widget.model != null) model = widget.model;

    Libs.modelValidationFunctions[widget.field] = () {
      if (widget.field != null) {
        var value = Dynamic(model).get(widget.field);
        ValidationDataRF data = ValidationRF.validate(widget.field, model, value);
        try {
          setState(() {
            Libs.modelValidationData[widget.field] = data;
          });
        } catch (w) {}

        return data?.isValid;
      }
      return false;
    };

    selectedValue(snapshot) {
      var defaultValue = "0";

      if (widget.field != null) {
        if (model != null) {
          var val = Dynamic(model).get(widget.field);
          defaultValue = (val == null) ? "0" : Dynamic(model).get(widget.field).toString();
        }
      } else {
        defaultValue = widget.value;
      }
      if (_selectedValue.isNotEmpty) defaultValue = _selectedValue;
      print('idx:$defaultValue');
      return defaultValue.toString();
    }

    simpleDropdown(snapshot) {
      var currentSelected = selectedValue(snapshot);
      print('current:selected : $currentSelected');
      return DropdownButton(
        onTap: () => FocusScope.of(context).requestFocus(FocusNode()),
        underline: Padding(padding: EdgeInsets.all(0)),
        isExpanded: true,
        items: snapshot as List<DropdownMenuItem<String>>,
        value: selectedValue(snapshot),
        style: TextStyle(fontSize: widget.textSize ?? 2.2.h, color: widget.textColor ?? Colors.black),
        onChanged: (value) {
          var index = (snapshot as List).indexWhere((element) => (element.value == value));
          if (widget.onChange != null) widget.onChange(index, value);
          if (widget.field != null) {
            if (model != null) Dynamic(model).set(widget.field, value);
            if (Libs.modelValidationData[widget.field] != null) Libs.modelValidationData[widget.field].isValid = true;
          }
          _selectedValue = value;
          setState(() {});
        },
      );
    }

    var templateDefaultControlBG; // future link with stock object
    StyleRF styleRF = Libs.config.app?.theme?.controlBackground ?? templateDefaultControlBG ?? widget.backColor;

    searchableDropdown(snapshot) {
      return Container(
        color: styleRF.backgroundColor,
        child: Stack(
          children: [
            SearchableDropdown.single(
              onSearchDialogOpen: () {},
              underline: Padding(padding: EdgeInsets.all(0)),
              items: snapshot as List<DropdownMenuItem<String>>,
              value: selectedValue(snapshot),
              isCaseSensitiveSearch: false,
              searchFn: (String keyword, items) {
                List<int> ret = List<int>();
                if (keyword != null && items != null && keyword.isNotEmpty) {
                  keyword.split(" ").forEach((k) {
                    int i = 0;
                    items.forEach((item) {
                      if (k.isNotEmpty && (item.child.data.toString().toLowerCase().contains(k.toLowerCase()))) {
                        ret.add(i);
                      }
                      i++;
                    });
                  });
                }
                if (keyword.isEmpty) {
                  ret = Iterable<int>.generate(items.length).toList();
                }
                return (ret);
              },
              hint: "Select an Item",
              searchHint: "Select an Item",
              onChanged: (value) {
                var index = (snapshot as List).indexWhere((element) => (element.value == value));
                if (widget.onChange != null) widget.onChange(index, value);
                Dynamic(model).set(widget.field, value);
                if (Libs.modelValidationData[widget.field] != null) {
                  Libs.modelValidationData[widget.field].isValid = true;
                }
                setState(() {});
              },
              isExpanded: true,
            ),
          ],
        ),
      );
    }

    //'f7f7f7'.toColor,
    var futureDropdown = FutureBuilder(
      future: (widget.items is Future) ? widget.items : null,
      builder: (context, snapshot) {
        switch (snapshot.connectionState) {
          case ConnectionState.waiting:
            return Text('Loading....');
          default:
            if (snapshot.hasError) {
              return Text('Error: ${snapshot.error}');
            } else {
              _selectedIndex = 0;
              return (widget.dropdownType == DropdownType.simple)
                  ? simpleDropdown(snapshot.data)
                  : searchableDropdown(snapshot.data);
            }
        }
      },
    );

    Color color = Libs.config.app?.theme?.controlBackground?.backgroundColor ?? widget.backColor;

    switch (widget.style) {
      case DropdownStyle.normal:
        return Container(
          child: Column(
            crossAxisAlignment: CrossAxisAlignment.start,
            children: <Widget>[
              Container(
                padding: EdgeInsets.only(left: 10, right: 10),
                decoration: BoxDecoration(
                    color: color, //shape: BoxShape.circle,
                    borderRadius: BorderRadius.all(Radius.circular(4))),
                child: (widget.items is Future)
                    ? futureDropdown
                    : (widget.dropdownType == DropdownType.simple)
                        ? simpleDropdown(widget.items)
                        : searchableDropdown(widget.items),
              ),
              ValidationRF.showErrorMessage(widget.field)
            ],
          ),
        );
        break;

      case DropdownStyle.fluid:
        return Container(
          margin: EdgeInsets.only(left: 5, right: 5),
          child: Column(
            crossAxisAlignment: CrossAxisAlignment.start,
            children: <Widget>[
              Container(
                padding: EdgeInsets.only(left: 10, right: 10),
                decoration: BoxDecoration(
                    color: Colors.white, //shape: BoxShape.circle,
                    boxShadow: [
                      BoxShadow(
                        color: Colors.grey.withOpacity(0.4),
                        blurRadius: 3.0,
                        spreadRadius: 2.0,
                      ),
                    ],
                    borderRadius: BorderRadius.all(Radius.circular(10))),
                child: (widget.items is Future)
                    ? futureDropdown
                    : (widget.dropdownType == DropdownType.simple)
                        ? simpleDropdown(widget.items)
                        : searchableDropdown(widget.items),
              ),
              ValidationRF.showErrorMessage(widget.field)
            ],
          ),
        );
        break;
    }
    return Container();
  }
}
