import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:catcher/catcher.dart';
import 'package:flutter_localizations/flutter_localizations.dart';
import 'package:flutter_rating_bar/flutter_rating_bar.dart';
import 'package:flutter_sticky_header/flutter_sticky_header.dart';
import 'package:flutter_xlider/flutter_xlider.dart';
import 'package:intl/intl.dart';
import 'package:rapidfire/Annotation.dart' as annotation;
import 'package:rapidfire/Config.dart';
import 'package:rapidfire/ValidationRF.dart';
import 'package:rapidfire/LocalizationProviderRF.dart';
import 'package:rapidfire/store/StoreRF.dart';
import 'package:rapidfire/store/modern/onboarding/Intro.dart';
import 'package:rapidfire/store/modern/onboarding/Splash.dart';
import 'package:rapidfire/uiux/intro/Models/page_view_model.dart';
import 'package:font_awesome_flutter/font_awesome_flutter.dart';
import 'package:sliver_fab/sliver_fab.dart';
import 'package:sizer/sizer.dart';
import 'package:rapidfire/RapidFire.dart';
import 'dart:convert';
// import 'package:keyboard_visibility/keyboard_visibility.dart';

import '../Db.dart';
import '../Dynamic.dart';
import '../GlobalRF.dart';
import '../LocalizationRF.dart';
import '../RapidFire.dart';
import '../uiux/StyleRF.dart';
import 'AlertRF.dart';
import 'DrawerRF.dart';
import 'WidgetStylesRF.dart';
import '../GlobalScopeRF.dart';
import '../Extensions.dart';
import 'package:rapidfire/SystemRF.dart';
import 'package:graphic/graphic.dart' as graphic;
import '../StackRF.dart';

export '../uiux/DialogRF.dart';
export 'package:rapidfire/uiux/widgets/AnimatedContainerRF.dart';
export 'StyleRF.dart';
export 'package:rapidfire/uiux/widgets/ButtonRF.dart';
export 'package:font_awesome_flutter/font_awesome_flutter.dart';
export 'package:rapidfire/uiux/DropdownListRF.dart';
export 'package:sizer/sizer.dart';
export 'package:rapidfire/store/StoreRF.dart';

//export 'package:rapidfire/uiux/widgets/ListViewRFOLD.dart';
export 'package:rapidfire/uiux/widgets/ListViewRF.dart';

//import 'package:rapidfire/uiux/swl/sweetalert.dart';

//#region RapidFireApp
class RapidFireAppParentRF extends InheritedWidget {
  Object initialModel;
  DataSourceRF dataSource;
  bool keyboardVisible;

  RapidFireAppParentRF({
    Key key,
    this.dataSource,
    this.keyboardVisible = false,
    this.initialModel,
    Widget child,
  }) : super(key: key, child: child);

  @override
  bool updateShouldNotify(RapidFireAppParentRF oldWidget) {
    return oldWidget.dataSource != dataSource || oldWidget.keyboardVisible != keyboardVisible;
  }
}

class RapidFireAppRF extends StatefulWidget {
  final dynamic child;

  RapidFireAppRF(this.child);

  static RapidFireAppParentRF of(BuildContext context) =>
      context.inheritFromWidgetOfExactType(RapidFireAppParentRF) as RapidFireAppParentRF;

  static dispose() {}

  @override
  State<StatefulWidget> createState() {
    return _RapidFireAppRF();
  }
}

class _RapidFireAppRF extends State<RapidFireAppRF> {
  LanguageDelegate _localeLanguageOverrideDelegate;

  @override
  void initState() {
    super.initState();
    var lang = rf.pref.getString("lang") ?? 'en';
    _localeLanguageOverrideDelegate = new LanguageDelegate(Locale(lang, ''));
    LanguageRF.loadLanguage = onSetLanguage;
  }

  onSetLanguage(Locale locale, Function onChanged) {
    setState(() {
      _localeLanguageOverrideDelegate = new LanguageDelegate(locale);
      rf.pref.setString('lang', locale.languageCode);
      LanguageRF.onLoadComplete = onChanged;
    });
  }

  @override
  Widget build(BuildContext context) {
    return RapidFireAppParentRF(
      child: LayoutBuilder(builder: (context, constraints) {
        return OrientationBuilder(
          builder: (context, orientation) {
            SizerUtil().init(constraints, orientation);
            return MaterialApp(
              navigatorKey: Catcher.navigatorKey,
              debugShowCheckedModeBanner: false,
              theme: Libs.config.app.themeData.customThemeData(ThemeDataTypeRF.light),
              darkTheme: Libs.config.app.themeData.customThemeData(ThemeDataTypeRF.dark),
              localizationsDelegates: [
                _localeLanguageOverrideDelegate,
                GlobalMaterialLocalizations.delegate,
                GlobalWidgetsLocalizations.delegate,
              ],
              supportedLocales: LanguageRF.supportedLocales(),
              home: widget.child,
            );
          },
        );
      }),
    );
  }
}
//#endregion RapidFireApp

//#region Onboard
enum OnboardItemsRF { splash, language, intro, login, otp, setup, home }

class AppOnboardRF {
  Widget language;
  dynamic splash;
  dynamic intro;
  Widget login;
  Widget otp;
  Widget initialSetup;
  dynamic home;

  // dynamic drawer;

  AppOnboardRF({
    this.language,
    this.splash,
    @required this.intro,
    this.login,
    this.otp,
    this.initialSetup,
    @required this.home, //   this.drawer,
  });

  static int currentStep = 0;
  static AppOnboardRF onboardItems;
  static List<dynamic> onboardItemList = List();
  static bool isLanguageLoaded = false;

  static loadOnboardItems(AppOnboardRF items) {
    onboardItems = items;
    onboardItemList.add(onboardItems.splash);
    onboardItemList.add(onboardItems.language);
    onboardItemList.add(onboardItems.intro);
    onboardItemList.add(onboardItems.login);
    onboardItemList.add(onboardItems.otp);
    onboardItemList.add(onboardItems.initialSetup);
    onboardItemList.add(onboardItems.home);
  }

  next() {
    for (int i = currentStep; i < onboardItemList.length; i++) {
      if (onboardItemList[i] != null) {
        currentStep++;
        return onboardItemList[i];
      } else {
        currentStep++;
      }
    }
  }

  static onboardWidget(source) {
    if (source is SplashRF)
      return Splash(splash: source);
    else if (source is List<IntroPageViewRF>)
      return Intro(pages: source);
    else
      return source;
  }

  static addItem(
    Configuration _configuration,
    OnboardItemsRF item,
  ) {
    TemplateRF clientTemplate = _configuration.app.template;
    if (_configuration.app.onboard is AppOnboardRF) {
      switch (item) {
        case OnboardItemsRF.splash:
          return onboardWidget(_configuration.app.onboard.splash) ??
              StoreRF.onboard.splash(_configuration, template: clientTemplate);
        case OnboardItemsRF.language:
          return !(_configuration.app.onboard.language is EmptyRF)
              ? onboardWidget(_configuration.app.onboard.language) ??
                  StoreRF.onboard.language(_configuration, template: clientTemplate)
              : null;
        case OnboardItemsRF.intro:
          return !(_configuration.app.onboard.intro is EmptyRF)
              ? onboardWidget(_configuration.app.onboard.intro)
              : null;
        case OnboardItemsRF.login:
          return !(_configuration.app.onboard.login is EmptyRF)
              ? onboardWidget(_configuration.app.onboard.login) ??
                  StoreRF.onboard.login(_configuration, template: clientTemplate)
              : null;
        case OnboardItemsRF.otp:
          return !(_configuration.app.onboard.otp is EmptyRF)
              ? onboardWidget(_configuration.app.onboard.otp) ??
                  StoreRF.onboard.otp(_configuration, template: clientTemplate)
              : null;
        case OnboardItemsRF.setup:
          return !(_configuration.app.onboard.initialSetup is EmptyRF)
              ? onboardWidget(_configuration.app.onboard.initialSetup) ??
                  StoreRF.onboard.setup(_configuration, template: clientTemplate)
              : null;
        case OnboardItemsRF.home:
          assert(_configuration.app.onboard.home != null, 'AppOnboardRF requires Home');
          return _configuration.app.onboard.home;
      }
    } else {
      switch (item) {
        case OnboardItemsRF.splash:
          return onboardWidget(_configuration.app.onboard.customOnboard().splash) ??
              StoreRF.onboard.splash(_configuration, template: clientTemplate);
        case OnboardItemsRF.language:
          return !(_configuration.app.onboard.customOnboard().language is EmptyRF)
              ? onboardWidget(_configuration.app.onboard.customOnboard().language) ??
                  StoreRF.onboard.language(_configuration, template: clientTemplate)
              : null;
        case OnboardItemsRF.intro:
          return !(_configuration.app.onboard.customOnboard().intro is EmptyRF)
              ? onboardWidget(_configuration.app.onboard.customOnboard().intro)
              : null;
        case OnboardItemsRF.login:
          return !(_configuration.app.onboard.customOnboard().login is EmptyRF)
              ? onboardWidget(_configuration.app.onboard.customOnboard().login) ??
                  StoreRF.onboard.login(_configuration, template: clientTemplate)
              : null;
        case OnboardItemsRF.otp:
          return !(_configuration.app.onboard.customOnboard().otp is EmptyRF)
              ? (onboardWidget(_configuration.app.onboard.customOnboard().otp) ??
                  StoreRF.onboard.otp(_configuration, template: clientTemplate))
              : null;
        case OnboardItemsRF.setup:
          return !(_configuration.app.onboard.customOnboard().initialSetup is EmptyRF)
              ? onboardWidget(_configuration.app.onboard.customOnboard().initialSetup) ??
                  StoreRF.onboard.setup(_configuration, template: clientTemplate)
              : null;
        case OnboardItemsRF.home:
          assert(_configuration.app.onboard.customOnboard().home != null, 'AppOnboardRF requires Home');
          return _configuration.app.onboard.customOnboard().home;
      }
    }
  }
}
//#endregion Onboard

//#region PageRF
class PageParentRF extends InheritedWidget {
  final Object model;
  final dynamic instance;
  FlexBarRF flex;

  PageParentRF({
    Key key,
    this.model,
    this.instance,
    this.flex,
    Widget child,
  }) : super(key: key, child: child);

  @override
  bool updateShouldNotify(PageParentRF oldWidget) {
    return oldWidget.model != model || oldWidget.instance != instance || oldWidget.flex != flex;
  }
}

enum PageType { none, home, entry, list, sync, fetch }

class PageRF extends StatefulWidget {
  final String title;
  final bool titleVisible;
  final String flexTitle;
  final bool flexTitleVisible;
  final AppBar appBar;
  final Widget appBarBottom;
  final Widget drawer;
  final Widget drawerIcon;
  final AppBar footer;
  final PageType pageType;
  final OutlineRF outline;
  final Object model;
  final Function onCallerBack;
  final FlexBarRF flex;
  final stickyHeader;
  final Widget body;
  final Widget fab; //fabEx + on click
  final FloatingActionButtonAnimator fabEx; //fabEx + on click
  final Widget filter;
  final Function onInit;
  final Function onBack;
  final Color backgroundColor;
  final bool disabledPageScroll;
  final List<Color> osBarColors;
  TemplateRF template;

  PageRF({
    @required this.title,
    this.titleVisible = false,
    this.flexTitle = '', //will go to flex
    this.flexTitleVisible = true,
    this.pageType,
    this.appBar, //New
    this.appBarBottom, //New
    this.drawer,
    this.drawerIcon,
    this.footer, //new
    this.outline, //?? no need
    this.model,
    this.onCallerBack,
    this.flex,
    this.stickyHeader,
    this.body,
    this.fab,
    this.fabEx,
    this.filter, // no need , can go to sticket header
    this.onInit,
    this.onBack,
    this.backgroundColor = Colors.white,
    this.disabledPageScroll = false,
    this.osBarColors,
    this.template,
  }) : assert(title != null);

  static DataSourceRF dataSourceRefreshOnBack; // remove
  static Function onCallerPageBack; // remove
  static Map<String, Function> onCallerRefreshMap = Map(); // remove
  static Map<String, List<DataSourceRF>> dataSourcesCallerRefreshMap = Map(); // remove

  static dynamic currentPageClassName;
  static dynamic currentDialogContext;
  static bool keyboardOpened;
  static double pageViewInsets;

  static PageParentRF of(BuildContext context) {
    return context.inheritFromWidgetOfExactType(PageParentRF) as PageParentRF;
  }

  // static refresh(BuildContext context) => (context as Element).reassemble();
  static refresh(dynamic instance) {
    if (instance is StatefulElement) {
      instance.reassemble();
    }
  }

  static bool isValidModel(BuildContext context, dynamic model, {String message, Function validator}) {
    bool valid = true;

    var modelMembers = Dynamic(model).modelMembers;

    // Annotation validations
    for (var item in Libs.modelValidationFunctions.entries) {
      if (modelMembers[item.key] != null) {
        var ret = item.value.call();
        if (ret != null) {
          valid = ret;
          if (!valid) Libs.modelValidationData[item.key].isValid = false;
        }
      }
    }
    if (!valid) {
      rf.msg(context, message ?? 'Data save was unsuccessful due to missing mandatory fields!');
      return valid;
    }

    // Global config validations
    for (var element in Libs.config?.model?.validators) {
      valid = element(context);
      var methodName = Dynamic.methodName(element);
      if (!valid) {
        return valid;
      }
    }

    // Local config validations
    if (validator != null) {
      String msg = validator.call() ?? '';
      if (msg.isNotEmpty) {
        rf.msg(context, msg ?? 'Data save was unsuccessful due some error!');
        return false;
      }
    }

    return valid;
  }

  static bool executeValidators(BuildContext context, String methodName, {Function validateHandler}) {
    bool valid = true;

    for (var element in Libs.config?.model?.validators) {
      var functionName = Dynamic.methodName(element);

      if (methodName == null || methodName.isEmpty) {
        valid = element(context);
      } else {
        if (functionName == methodName) {
          valid = element(context);
        }
      }

      if (!valid) {
        if (validateHandler != null) validateHandler(functionName);
      }
    }
    return valid;
  }

  static willPop(context, {model}) {
    if (model != null) {
      if (Dynamic(model).changed(RapidFireAppRF.of(context).initialModel)) {
        rf.msg(
          context,
          'Save change pending, do you want to quite anyway?',
          icon: Icons.error,
          actionText: 'Quit',
          action: () => Navigator.pop(context),
        );
      } else {
        if (Navigator.canPop(context)) {
          Navigator.pop(context);
        } else {
          //SystemChannels.platform.invokeMethod('SystemNavigator.pop');
          SystemNavigator.pop();
        }
      }
    } else {
      if (Navigator.canPop(context)) {
        Navigator.pop(context);
      } else {
        //SystemChannels.platform.invokeMethod('SystemNavigator.pop');
        SystemNavigator.pop();
      }
    }
    return Future.value(true);
  }

  @override
  State<StatefulWidget> createState() => _PageRF(osBarColors);
}

class _PageRF extends State<PageRF> {
  BuildContext _pageContext;
  List<Color> _colors;

  dynamic modelPKValue, initialModelPKValue;

  _PageRF(List<Color> colors) {
    _colors = colors;
  }

  @override
  void initState() {
    _pageContext = context;
    widget.template = widget.template ?? TemplateRF(TemplateNameRF.modern);

    PageRF.dataSourceRefreshOnBack = null;
    Libs.modelValidationData.clear();
    Libs.modelValidationFunctions.clear();

    SystemRF.os.setOsBarColors(_colors);

    WidgetsBinding.instance.addPostFrameCallback((_) {
      if (widget.onInit != null) widget.onInit(context);
    });

    RapidFire.updateGlobalErrorLanguage();

    super.initState();
  }

  @override
  void didChangeDependencies() async {
    if (widget.model != null) {
      ValidationRF.validationAnnotations = Dynamic(widget.model).annotationData(annotation.Check);

      var _initialModel = Dynamic(widget.model).copy();

      if (RapidFireAppRF.of(context).initialModel == null) {
        RapidFireAppRF.of(context).initialModel = _initialModel;
        initialModelPKValue = Db().getPrimaryKeyValue(_initialModel);
      } else {
        if (widget.model.runtimeType.toString() != RapidFireAppRF.of(context).initialModel.runtimeType.toString()) {
          RapidFireAppRF.of(context).initialModel = _initialModel;
          initialModelPKValue = Db().getPrimaryKeyValue(_initialModel);
        }
      }
    }
    super.didChangeDependencies();
  }

  @override
  void dispose() {
    //if (PageRF.callerStack.length == 1) PageRF.callerStack.clearStack();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    PageRF.pageViewInsets = MediaQuery.of(context).viewInsets.bottom;
    if (PageRF.currentDialogContext != null) {
      //(context as Element).reassemble();
    }
    Libs.context = context;

    if (widget.model != null) {
      if (initialModelPKValue == null) {
        modelPKValue = Db().getPrimaryKeyValue(widget.model);
        if (modelPKValue != null && modelPKValue != "") {
          RapidFireAppRF.of(context).initialModel = Dynamic(widget.model).copy();
        }
      }
    }

    // OutlineRF.loadPageOutline(widget.template);

    var page = PageParentRF(
      instance: this,
      model: widget.model,
      flex: widget.flex,
      child: Scaffold(
        backgroundColor: widget.backgroundColor ?? Colors.white,
        //appBar: Libs.config.app.outline.customOutline().appBar ?? StoreRF.implementation(template).page.outline.customOutline().appBar,
        //appBar: appBar,
        // OutlineRF.getAppBar(
        //   context,
        //   widget.outline,
        //   widget.title,
        //   widget.flex,
        //   widget.outline?.stickyHeaderPadding,
        //   widget.template,
        // ),
        drawer: widget.drawer,
        //OutlineRF.getAppDrawer(widget.outline, widget.template),
        body: WillPopScope(
          onWillPop: widget.onBack ??
              () {
                PageRF.willPop(context, model: widget.model);
                return Future.value(true);
              },
          child: pageContent(
            widget.flex,
            widget.title,
            widget.titleVisible,
            widget.flexTitle,
            widget.flexTitleVisible,
            widget.stickyHeader,
            widget.disabledPageScroll,
          ),
        ),
        floatingActionButton: widget.fab,
        floatingActionButtonAnimator: widget.fabEx,
        bottomNavigationBar: widget.outline?.footer,
        //PageOutlineRF.getAppBottomBar(widget.outline, widget.theme),
      ),
    );

    return page;
  }

  sliverAppBar(BuildContext context, Widget leading, FlexBarRF flex, String title) {
    var _expandedHeight = flex?.height ?? 0.0;
    var titleCenter = true;
    var top = 0.0;

    var appBar = widget.appBar ??
        Libs.config.page.appBar
            .customAppBar(widget.pageType, context, widget.model) ?? // .customAppBar.customAppBar(widget.pageType) ??
        StoreRF.implementation(widget.template).page.appBar.customAppBar(widget.pageType, context, widget.model);

    flexPanel() {
      return LayoutBuilder(builder: (BuildContext context, BoxConstraints constraints) {
        top = constraints.biggest.height;
        if (top <= 80) {
          titleCenter = false;
        } else {
          titleCenter = true;
        }
        return FlexibleSpaceBar(
          centerTitle: titleCenter,
          collapseMode: CollapseMode.parallax,
          stretchModes: [StretchMode.fadeTitle],
          title: AnimatedOpacity(
            duration: Duration(milliseconds: 100),
            opacity: top == 80.0 ? 1.0 : 0.0,
            child: Text(
              title, //widget.flex.title,
              style: TextStyle(fontSize: 20.0, color: Colors.black),
            ),
          ),
          background: Container(
            margin: EdgeInsets.only(top: 25),
            color: flex.color,
            child: Stack(
              children: <Widget>[
                flex.background,
                Positioned(
                  left: 15,
                  bottom: 15,
                  child: Text(
                    flex.label ?? '',
                    style: TextStyle(fontSize: 20.0, color: Colors.black, fontWeight: FontWeight.bold),
                  ),
                ),
              ],
            ),
          ),
        );
        //Image.network(widget.flex.background, fit: BoxFit.cover));
      });
    }

    return SliverAppBar(
      iconTheme: IconThemeData(color: Colors.black),
      leading: appBar.leading,
      title: flex == null
          ? TextRF(
              title,
              size: 18,
              color: Colors.black87,
            )
          : TextRF(''),
      expandedHeight: _expandedHeight,
      floating: true,
      backgroundColor: appBar.backgroundColor,
      actions: appBar.actions,
      pinned: true,
      elevation: appBar.elevation ?? 0.0,
      forceElevated: (appBar.elevation == null) ? false : true,
      bottom: widget.appBarBottom,
      flexibleSpace: flex != null ? flexPanel() : null,
    );
  }

  pageContent(FlexBarRF flex, String title, bool titleVisible, String flexTitle, bool flexTitleVisible,
      Widget stickyHeader, bool disablePageScroll) {
    var content = CustomScrollView(
      physics: disablePageScroll ? NeverScrollableScrollPhysics() : null,
      slivers: <Widget>[
        sliverAppBar(context, widget?.outline?.appBar?.leading, flex, title),
        SliverStickyHeader(
          header: Container(
            height: stickyHeader == null ? 0.0 : (flex == null ? 64 : 56),
            color: Colors.white,
            padding: EdgeInsets.only(top: 0.0),
            alignment: Alignment.center,
            child: Padding(padding: EdgeInsets.only(top: (flex == null ? 10 : 0)), child: stickyHeader),
          ),
          sliver: SliverList(delegate: SliverChildListDelegate([widget.body])),
        ),
      ],
    );

    if (Libs.config.page.scrollPhysics == null) {
      return ScrollConfiguration(
        behavior: BounceScrollBehavior(),
        child: content,
      );
    } else {
      if (Libs.config.page.scrollPhysics == ScrollPhysicsRF.bounce) {
        return ScrollConfiguration(
          behavior: BounceScrollBehavior(),
          child: content,
        );
      } else {
        return content;
      }
    }
  }
}

class FlexBarRF {
  var label;
  var background;
  var backgroundHeight;
  var color;
  var height;

  // var stickyHeader;
  FlexBarRF({
    this.label,
    this.background,
    this.color = Colors.white,
    this.height = 250.0,
    this.backgroundHeight = 300,
  });
}
//#endregion PageRF

//#region PageOutline
enum OutlineDisableItems { header, headerPanel, pageFooter }

class OutlineRF {
  AppBarRF appBar;
  Widget stickyHeader;
  Widget footer;
  EdgeInsets stickyHeaderPadding;
  TemplateRF template;

//  style -> bar ->back color, font, elevation
  //pageBar sub class=> leading, actions, drawer, , , ?style?, -disabled, -title?

  OutlineRF({
    this.appBar,
    this.stickyHeader,
    this.footer,
    this.stickyHeaderPadding,
    this.template,
  });

  static OutlineRF implemented = OutlineRF(); //client, stock,

  static loadPageOutline(template) {
    implemented.appBar = Libs.config.app.outline.customOutline().appBar ??
        StoreRF.implementation(template).page.outline.customOutline().appBar;

    implemented.appBar?.drawer = Libs.config.app.outline.customOutline().appBar?.drawer ??
        StoreRF.implementation(template).page.outline.customOutline().appBar?.drawer;

    implemented.appBar?.actions = Libs.config.app.outline.customOutline().appBar?.actions ??
        StoreRF.implementation(template).page.outline.customOutline().appBar?.actions;

    implemented.stickyHeader = Libs.config.app.outline.customOutline().stickyHeader ??
        StoreRF.implementation(template).page.outline.customOutline().stickyHeader;

    implemented.footer = Libs.config.app.outline.customOutline().footer ??
        StoreRF.implementation(template).page.outline.customOutline().footer;
  }

  static getAppBar(
    BuildContext context,
    OutlineRF local,
    String pageTitle,
    FlexBarRF flex,
    EdgeInsets pageHeaderPanelPadding,
    TemplateRF template,
  ) {
    OutlineRF _source = OutlineRF();
    var appBar;

    // _source.pageHeader = local?.pageHeader ?? implemented.pageHeader;
    _source.appBar = AppBarRF(
      leading: local?.appBar?.leading ?? implemented.appBar.leading,
      drawer: local?.appBar?.drawer ?? implemented.appBar.drawer,
      actions: local?.appBar?.actions ?? implemented.appBar.actions,
      backgroundColor: local?.appBar?.backgroundColor ?? implemented.appBar.backgroundColor,
      foregroundColor: local?.appBar?.foregroundColor ?? implemented.appBar.foregroundColor,
      elevation: local?.appBar?.elevation ?? implemented.appBar.elevation,
    );
    _source.stickyHeader = local?.stickyHeader ?? implemented.stickyHeader;

    assert(pageTitle != null, "Page title is missing.");

    customAppBar() {
      AppBar pageHeader;
      var _headerPanel = _source?.stickyHeader;
      var pageBottom = PreferredSize(
        child: Container(
          color: Colors.transparent,
          height: 50.0,
          width: ScreenRF.width(context),
          child: Padding(
            padding:
                EdgeInsets.only(top: pageHeaderPanelPadding?.top ?? 5, bottom: pageHeaderPanelPadding?.bottom ?? 5),
            child: _headerPanel is Widget ? _headerPanel : (_headerPanel as TabBarRF).toTabBar,
          ),
        ),
        preferredSize: Size.fromHeight(50.0),
      );

      if (_source?.appBar is AppBarRF) {
        pageHeader = AppBar(
          titleSpacing: 0,
          actionsIconTheme: IconThemeData(color: (_source?.appBar as AppBarRF).foregroundColor ?? Colors.white),
          elevation: (_source?.appBar as AppBarRF).elevation,
          backgroundColor: (_source?.appBar as AppBarRF).backgroundColor,
          leading: (_source?.appBar as AppBarRF).leading is EmptyRF ? null : (_source?.appBar as AppBarRF).leading,
          title: Text(pageTitle, style: TextStyle(color: (_source?.appBar as AppBarRF).foregroundColor)),
          actions: (_source?.appBar as AppBarRF).actions,
          bottom: _headerPanel == null ? _headerPanel : pageBottom,
        );
        return pageHeader;
      } else if (_source?.appBar is AppBar) {
        pageHeader = _source?.appBar as AppBar; //ImroseX
      } else if (_source?.appBar is EmptyRF) {
        return null;
      }

      return AppBar(
        leading: (pageHeader)?.leading is EmptyRF ? null : (pageHeader)?.leading,
        automaticallyImplyLeading: (pageHeader)?.automaticallyImplyLeading ?? true,
        title: Text(pageTitle),
        actions: (pageHeader)?.actions,
        flexibleSpace: (pageHeader)?.flexibleSpace,
        bottom: PreferredSize(
            child: Container(
              color: Colors.transparent,
              height: 50.0,
              width: 800,
              child: _headerPanel is Widget ? _headerPanel : (_headerPanel as TabBarRF).toTabBar,
            ),
            preferredSize: Size.fromHeight(50.0)),
        elevation: (pageHeader)?.elevation,
        shape: (pageHeader)?.shape,
        //backgroundColor: (pageHeader)?.backgroundColor,
        brightness: (pageHeader)?.brightness,
        iconTheme: (pageHeader)?.iconTheme,
        actionsIconTheme: (pageHeader)?.actionsIconTheme,
        textTheme: (pageHeader)?.textTheme,
        primary: (pageHeader)?.primary ?? true,
        centerTitle: (pageHeader)?.centerTitle,
        excludeHeaderSemantics: (pageHeader)?.centerTitle ?? false,
        titleSpacing: (pageHeader)?.titleSpacing ?? NavigationToolbar.kMiddleSpacing,
        toolbarOpacity: (pageHeader)?.toolbarOpacity ?? 1.0,
        bottomOpacity: (pageHeader)?.bottomOpacity ?? 1.0,
      );
    }

    // (local?.pageHeader as AppBar).leading = (local?.pageHeader as PageOutlineRF).headerPanel.;

    if (flex == null) {
      appBar = customAppBar();
    }
    return appBar;
  }

  static getAppDrawer(OutlineRF local, TemplateRF template) {
    OutlineRF _source = OutlineRF(
      appBar: AppBarRF(
        //     leading: (local?.pageHeader as PageHeaderRF)?.leading ??
        //            (implemented.pageHeader as PageHeaderRF).leading,
//        actions: local?.pageHeader is EmptyRF
//            ? null
//            : (local?.pageHeader as PageHeaderRF)?.actions ??
//                (implemented.pageHeader as PageHeaderRF).actions,
        drawer: local?.appBar is EmptyRF
            ? null
            : (local?.appBar as AppBarRF)?.drawer ?? (implemented.appBar as AppBarRF).drawer,
      ),
    );
    var drawer = (_source.appBar as AppBarRF).drawer;
    return drawer is EmptyRF ? null : drawer;
  }

//default+local
//static getAppTabBar(){}

  //default check==>>
  static getAppBottomBar(OutlineRF local, TemplateRF template) {
    OutlineRF _source = OutlineRF();
    _source.footer = local?.footer ?? implemented.footer ?? implemented.footer;
    return _source.footer;
  }
}

class AppBarRF {
  Widget leading;
  List<Widget> actions;
  dynamic drawer;
  double elevation;
  Color backgroundColor;
  Color foregroundColor;

  AppBarRF({
    this.leading,
    this.actions,
    this.drawer,
    this.elevation,
    this.backgroundColor,
    this.foregroundColor,
  });
}
//#endregion PageOutline

//#region EmptyRF
class EmptyRF extends StatelessWidget {
  static of(Type type, [bool list = false]) {
    if (type == DrawerItemRF) {
      if (list) {
        return List<DrawerItemRF>();
      } else {
        return DrawerItemRF(template: TemplateRF(null, version: 'EmptyRF'));
      }
    } else if (type == List) {
      return List<Widget>();
    }
  }

  @override
  Widget build(BuildContext context) => Container(height: 0, width: 0);
}
//#endregion EmptyRF

//#region TabBar
class TabBuilderRF {
  final List<Widget> views;
  final Color selectedLabelColor;
  final Color unselectedLabelColor;
  final TabBarIndicatorSize indicatorSize;
  final Decoration indicatorDecoration;
  final Function onTap;
  final TickerProvider tickerProvider;
  TabController controller;
  List<Widget> tabs;
  Function tabBuilder;
  Function _moveBefore;
  Function _moveAfter;

  moveAfter() {
    _moveAfter();
  }

  moveBefore() {
    _moveBefore();
  }

  TabBuilderRF({
    this.tabs,
    this.views,
    this.controller,
    this.selectedLabelColor,
    this.unselectedLabelColor,
    this.indicatorSize = TabBarIndicatorSize.tab,
    this.indicatorDecoration, // = const BoxDecoration(color: Colors.blue),
    this.onTap,
    this.tickerProvider,
    this.tabBuilder,
  }) {
    _moveBefore = moveBefore;
    _moveAfter = moveAfter;
    //var tabs1 = tabBuilder(0) as List<Widget>;
    if (tabBuilder != null) this.tabs = tabBuilder(0) as List<Widget>;
    this.controller = TabController(length: tabs.length, vsync: tickerProvider);
  }
}

class TabBarRF extends StatefulWidget {
  TabBuilderRF tab;
  TabBuilderRF view;

  TabBarRF({this.tab, this.view});

  @override
  State<StatefulWidget> createState() {
    return _TabBarViewRF();
  }
}

class _TabBarViewRF extends State<TabBarRF> {
  TabController tabController;
  int _selectedIndex = 0;
  int tabLength = 0;

  void initState() {
    if (widget.tab != null) {
      tabController = widget.tab.controller;
      tabLength = widget.tab.tabs.length;
    } else {
      tabController = widget.view.controller;
      tabLength = widget.view.tabs.length;
    }
    tabController.addListener(() {
      setState(() {
        _selectedIndex = tabController.index;
      });
    });
    super.initState();
  }

  __moveBefore() {
    if (_selectedIndex > 0) {
      setState(() => _selectedIndex--);
      tabController.animateTo(_selectedIndex);
    }
  }

  __moveAfter() {
    if (_selectedIndex < (tabLength - 1)) {
      setState(() => _selectedIndex++);
      tabController.animateTo(_selectedIndex);
    }
  }

  @override
  Widget build(BuildContext context) {
    dynamic tabItem;

//    tabController =
//        (tabController == null) ? TabController(length: 1, vsync: widget.tab.tickerProvider) : tabController;

    if (widget.tab != null) {
      widget.tab._moveBefore = __moveBefore;
      widget.tab._moveAfter = __moveAfter;
      if (widget.tab.tabBuilder != null) widget.tab.tabs = widget.tab.tabBuilder(_selectedIndex) as List<Widget>;
//      tabController = (tabController == null)
//          ? TabController(length: widget.tab.tabs.length, vsync: widget.tab.tickerProvider)
//          : tabController;
//      widget.tab.controller = tabController;

      tabItem = TabBar(
        labelColor: widget.tab.selectedLabelColor,
        unselectedLabelColor: widget.tab.unselectedLabelColor,
        indicatorSize: widget.tab.indicatorSize,
        isScrollable: true,
        indicator: widget.tab.indicatorDecoration,
        tabs: widget.tab.tabs,
        controller: tabController,
        onTap: (index) {
          setState(() => _selectedIndex = index);
          widget.tab.tabs = widget.tab.tabBuilder(index) as List<Widget>;
          if (widget.tab.onTap != null) widget.tab.onTap(_selectedIndex);
        },
      );
    } else if (widget.view != null) {
      tabItem = Container(
          height: 20,
          child: TabBarView(
            controller: widget.view.controller, //widget.view.controller,
            children: widget.view.views,
          ));
    }
    return tabItem;
  }
}

class TabDesignerRF extends StatefulWidget {
  int index;
  int selected;
  String label;
  Icon icon;
  Decoration decoration;

  TabDesignerRF(
    this.index,
    this.selected, {
    this.label,
    this.icon,
    this.decoration,
  });

  @override
  _TabDesignerRFState createState() => _TabDesignerRFState();
}

class _TabDesignerRFState extends State<TabDesignerRF> {
  @override
  Widget build(BuildContext context) {
    return StoreRF.scaffold().widgets.tabBar.getStyle(widget.index, widget.selected, label: widget.label);
  }
}
//#endregion TabBar

//#region ContainerRF
class ContainerRF extends StatefulWidget {
  final AlignmentGeometry alignment;

  const ContainerRF({
    this.alignment,
    this.padding,
    this.margin,
    this.height,
    this.decoration,
    this.child,
    this.template,
    this.style,
  });

  final EdgeInsetsGeometry padding;
  final EdgeInsetsGeometry margin;
  final double height;
  final Decoration decoration;
  final Widget child;
  final dynamic template;
  final StyleRF style;

  @override
  _ContainerRF createState() => _ContainerRF();
}

//IContainerStyleRF clint =
//    g.clientWidgetStyle?.everything ?? g.clientWidgetStyle.containerStyleRF;
class _ContainerRF extends State<ContainerRF> {
  @override
  Widget build(BuildContext context) {
    var currentTheme = widget.template ?? Libs.config.app.template ?? TemplateRF.get();
    IContainerRFStyle clint = Libs.config.app.widgetStyle.everything ?? Libs.config.app.widgetStyle.container;

    StyleContainerRF containerStyle =
        clint.containerRFStyle() ?? StoreRF.style(currentTheme).container.containerRFStyle();

    var alignment = widget.alignment ?? containerStyle.alignment;
    var padding = widget.padding ?? containerStyle.padding;
    var margin = widget.margin ?? containerStyle.margin;
    var decoration = widget.decoration ?? containerStyle.decoration;
    decoration = BoxDecoration(
      color: (decoration as BoxDecoration).color,
      image: (decoration as BoxDecoration).image,
      border: (decoration as BoxDecoration).border,
      borderRadius: (decoration as BoxDecoration).borderRadius,
      boxShadow: widget.style?.shadow ?? (decoration as BoxDecoration).boxShadow,
      gradient: (decoration as BoxDecoration).gradient,
      backgroundBlendMode: (decoration as BoxDecoration).backgroundBlendMode,
      shape: BoxShape.rectangle,
    );
    return Container(
      alignment: alignment,
      padding: padding,
      margin: margin,
      child: widget.child,
      decoration: decoration, //==>> remove StyleContainerRF to StyleRF
    );
  }
}
//#endregion ContainerRF

//#region TextRF
class TextRF extends StatefulWidget {
  final String text;
  final languageSource;
  final double size;
  final FontWeight weight;
  final Color color;
  final TextStyle style;
  final TextAlign align;
  final bool visible;
  final dynamic template;

  const TextRF(
    this.text, {
    this.languageSource,
    this.size,
    this.weight,
    this.color,
    this.visible = true,
    this.style,
    this.align = TextAlign.left,
    this.template,
  });

  @override
  _TextRF createState() => _TextRF();
}

//ITextStyle clintTextRF =
//    g.clientWidgetStyle?.everything ?? g.clientWidgetStyle.textStyleRF;
class _TextRF extends State<TextRF> {
  @override
  Widget build(BuildContext context) {
    var model = PageRF.of(context)?.model;

    ITextRFStyle clintTextRF = Libs.config.app.widgetStyle.everything ?? Libs.config.app.widgetStyle.text;
    StyleTextRF textStyle = clintTextRF.textRFStyle() ?? StoreRF.style(widget.template).text.textRFStyle();

    var templateDefaultControlBG; // future link with stock object
    StyleRF styleRF = Libs.config.app?.theme?.label ?? templateDefaultControlBG;

    var fontSize = widget.size ?? textStyle?.fontSize ?? styleRF.size;
    var fontWeight = widget.weight ?? textStyle?.fontWeight ?? styleRF.weight;
    var color = widget.color ?? textStyle?.color ?? styleRF.textColor;

    var _text = '';
    if (widget.languageSource != null) {
      _text = rf.lang.text(widget.languageSource, widget.text).replaceAll('not found', '').trim();
    } else {
      if (model != null) {
        _text = rf.lang.text(model.runtimeType.toString(), widget.text).replaceAll('not found', '').trim();
      } else {
        _text = widget.text;
      }
      /*if (_text == widget.text || _text == '') {
        _text = rf.lang.text('Common', widget.text).replaceAll('not found', '').trim();
      }*/
    }
    if (_text == widget.text || _text == '') {
      _text = rf.lang.text('Common', widget.text).replaceAll('not found', '').trim();
    }

    return Visibility(
        visible: widget.visible,
        child: Text(
          _text,
          textAlign: widget.align,
          style: widget.style ??
              TextStyle(
                fontSize: fontSize,
                fontWeight: fontWeight,
                color: color,
              ),
        ));
  }
}
//#endregion TextRF

//#region RowRF
class RowRF extends StatefulWidget {
  const RowRF(this.controls, [this.flex, this.x = 25, this.y = 10]);

  final dynamic controls;
  final double x;
  final double y;
  final List<int> flex;

  @override
  _RowRF createState() => _RowRF();
}

class _RowRF extends State<RowRF> {
  @override
  Widget build(BuildContext context) {
    int count = 1;
    var widgets = List<Widget>();

    for (var item in widget.controls) {
      var _flex;
      if (widget.flex != null) {
        _flex = widget?.flex[count - 1];
      }
      var ctl = Expanded(child: item, flex: _flex ?? 100);
      //print(_flex.toString());
      widgets.add(ctl);
      if (widget.controls.length != count) widgets.add(SizedBox(width: widget.y));
      count++;
    }
    if (widget.x != 0) {
      widgets.add(Container(
        height: widget.x,
      ));
    }

    return Row(
      mainAxisAlignment: MainAxisAlignment.end,
      //mainAxisAlignment: MainAxisAlignment.start,
      mainAxisSize: MainAxisSize.max,
      crossAxisAlignment: CrossAxisAlignment.start,
      children: widgets,
    );
  }
}
//#endregion RowRF

//region TextFieldRF
enum ValidationTypeRF { required, range, max, min }

class TextFieldRF extends StatefulWidget {
  TextFieldRF({
    this.model,
    this.field,
    this.inputType,
    this.maxLength,
    this.maxLines,
    this.hintText,
    this.hint,
    this.caption,
    this.onSaved,
    this.validator,
    this.onFieldSubmitted,
    this.validator1,
    this.onChange,
    this.backColor = const Color(0x77f7f7f7),
  });

  final Object model;
  dynamic field;
  final TextInputType inputType;
  final int maxLength;
  final int maxLines;
  final String hintText;
  final String hint;
  final String caption;

  final FormFieldSetter<String> onSaved;
  final FormFieldValidator<String> validator;
  final ValueChanged<String> onFieldSubmitted;

  final Function validator1;
  final Function onChange;
  final Color backColor;

  @override
  _TextFormFieldRF createState() => _TextFormFieldRF();
}

class _TextFormFieldRF extends State<TextFieldRF> {
  TextEditingController controller = TextEditingController();
  FocusNode _focus = new FocusNode(canRequestFocus: false);

  @override
  void initState() {
    _focus.addListener(_onFocusChange);
    super.initState();
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
  }

  void _onFocusChange() {
    if (!_focus.hasFocus) {
      setState(() {
        if (Libs.modelValidationData[widget.field] != null) Libs.modelValidationData[widget.field].isValid = true;
        //controller.selection = TextSelection.fromPosition(TextPosition(offset: controller.text.length));
      });
    }
  }

  @override
  Widget build(BuildContext context) {
    var model = PageRF.of(context)?.model;

    if (widget.model != null) model = widget.model;

    if (widget.field != null) {
      Libs.modelValidationFunctions[widget.field] = () {
        if (widget.field != null) {
          var value = Dynamic(model).get(widget.field);
          ValidationDataRF data = ValidationRF.validate(widget.field, model, value);
          setState(() {
            Libs.modelValidationData[widget.field] = data;
          });
          return data?.isValid;
        }
        return false;
      };
    }

    var fieldValue = (model == null)
        ? null
        : (widget.field == null)
            ? null
            : Dynamic(model).get(widget.field);

    if (fieldValue != null) {
      controller..text = fieldValue.toString();
    }

    var templateDefaultControlBG; // future link with stock object
    StyleRF styleRF = Libs.config.app?.theme?.controlBackground ?? templateDefaultControlBG ?? widget.backColor;
    Color color = Libs.config.app?.theme?.controlBackground?.backgroundColor ?? widget.backColor;
    var textField = Column(
      crossAxisAlignment: CrossAxisAlignment.start,
      children: <Widget>[
        Container(
          decoration: BoxDecoration(
              color: color, //shape: BoxShape.circle,
              /* boxShadow: [
                BoxShadow(
                  color: Colors.grey.withOpacity(0.5),
                  blurRadius: 2.0,
                  spreadRadius: 2.0,
                ),
              ],*/
              borderRadius: BorderRadius.all(Radius.circular(4))),
          child: TextField(
            enabled: true,
            controller: controller,
            focusNode: _focus,
            maxLines: widget.maxLines,
            keyboardType: widget.inputType,
            //autofocus: false,
            onChanged: (input) {
              if (model != null) Dynamic(model).set(widget.field, input.trim());
              if (widget.onChange != null) widget.onChange(input);
            },
            decoration: new InputDecoration(
              fillColor: styleRF.backgroundColor,
              //'f7f7f7'.toColor,
              border: InputBorder.none,
              focusColor: Colors.transparent,
              filled: true,
              hintText: widget.hint,
              //hintStyle: new TextStyle(color: Colors.black54),
              //prefix: Text('   '), //suffix: Text('   '),
              //contentPadding: EdgeInsets.only(left: 0, right: 0, top: 0),
            ),
          ),
        ),
        ValidationRF.showErrorMessage(widget.field)
      ],
    );

    return textField;
  }
}
//endregion TextFieldRF

//region DatePickerRF
class DatePickerRF extends StatefulWidget {
  Object model;
  String field;
  DateTime maxDate;
  DateTime minDate;
  Color backColor;
  double widgetHeight;
  Function onChange;

  @override
  _DatePickerRFState createState() => _DatePickerRFState();

  DatePickerRF(
      {this.model,
      this.field,
      this.maxDate,
      this.minDate,
      this.backColor = const Color(0x77f7f7f7),
      this.widgetHeight = 48,
      this.onChange}) {
    if (maxDate == null) maxDate = DateTime(2025);
    if (minDate == null) minDate = DateTime(1970);
  }
}

class _DatePickerRFState extends State<DatePickerRF> {
  String dateValue = 'Select Date';

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
  }

  @override
  Widget build(BuildContext context) {
    var model = PageRF.of(context)?.model;

    if (widget.model != null) model = widget.model;

    if (widget.field != null) {
      Libs.modelValidationFunctions[widget.field] = () {
        if (widget.field != null) {
          var value = Dynamic(model).get(widget.field) as String;
          ValidationDataRF data = ValidationRF.validate(widget.field, model, value);
          setState(() {
            Libs.modelValidationData[widget.field] = data;
          });
          return data?.isValid;
        }
        return false;
      };
      var value = Dynamic(model).get(widget.field);
      if (value != null) dateValue = value.toString();
    }

    var templateDefaultControlBG; // future link with stock object
    StyleRF styleRF = Libs.config.app?.theme?.controlBackground ?? templateDefaultControlBG ?? widget.backColor;

    //height: 48,
    // 'f7f7f7'.toColor,
    var datePicker = Container(
      decoration: BoxDecoration(
        color: styleRF.backgroundColor, //shape: BoxShape.circle,
        borderRadius: BorderRadius.all(Radius.circular(4)),
      ),
      child: InkWell(
        child: Center(child: Text(dateValue, style: TextStyle(color: Colors.black, fontWeight: FontWeight.normal))),
        onTap: () async {
          FocusScope.of(context).requestFocus(FocusNode());
          var dateFormat = DateFormat('yyyy-MM-dd');
          DateTime initialDate = DateTime.now();
          initialDate = DateTime.parse(dateFormat.format(initialDate));
          final DateTime picked = await showDatePicker(
              context: context, initialDate: initialDate, firstDate: widget.minDate, lastDate: widget.maxDate);
          setState(() {
            dateValue = dateFormat.format(picked);
            if (widget.field != null) Dynamic(model).set(widget.field, dateValue);
            if (widget.onChange != null) widget.onChange(dateValue);
            if (Libs.modelValidationData[widget.field] != null) Libs.modelValidationData[widget.field].isValid = true;
          });
        },
      ),
    );

    return Column(
      crossAxisAlignment: CrossAxisAlignment.start,
      children: <Widget>[
        Container(constraints: BoxConstraints.expand(height: widget.widgetHeight), child: datePicker),
        ValidationRF.showErrorMessage(widget.field)
      ],
    );
  }
}
//endregion DatePickerRF

//region TimePickerRF
class TimePickerRF extends StatefulWidget {
  String field;

  @override
  _TimePickerRFState createState() => _TimePickerRFState();

  TimePickerRF({this.field});
}

class _TimePickerRFState extends State<TimePickerRF> {
  String timeValue = 'Select Time';
  DateTime maxDate;
  DateTime minDate;

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
  }

  @override
  Widget build(BuildContext context) {
    var model = PageRF.of(context)?.model;

    Libs.modelValidationFunctions[widget.field] = () {
      if (widget.field != null) {
        var value = Dynamic(model).get(widget.field) as String;
        ValidationDataRF data = ValidationRF.validate(widget.field, model, value);
        setState(() {
          Libs.modelValidationData[widget.field] = data;
        });
        return data?.isValid;
      }
      return false;
    };

    var value = Dynamic(model).get(widget.field);
    if (value != null) timeValue = value.toString();

    var timePicker = Container(
      height: 45,
      color: 'f7f7f7'.toColor,
      child: InkWell(
        child: Center(
            child: Text(
          timeValue,
          style: TextStyle(color: Colors.grey, fontWeight: FontWeight.bold),
        )),
        onTap: () async {
          FocusScope.of(context).requestFocus(FocusNode());
          var myFormat = DateFormat('yyyy-MM-dd');
          DateTime initalDate = DateTime.now();
          initalDate = DateTime.parse(myFormat.format(initalDate));
          final selectedTime = await showTimePicker(
            initialTime: TimeOfDay.now(),
            context: context,
          );
          setState(() {
            timeValue = selectedTime.toString();
            Dynamic(model).set(widget.field, timeValue);
            if (Libs.modelValidationData[widget.field] != null) Libs.modelValidationData[widget.field].isValid = true;
          });
        },
      ),
    );

    return Column(
      crossAxisAlignment: CrossAxisAlignment.start,
      children: <Widget>[
        Container(constraints: BoxConstraints.expand(height: 45), child: timePicker),
        ValidationRF.showErrorMessage(widget.field)
      ],
    );
  }
}
//endregion TimePickerRF

//region SliderRF
class SliderRF extends StatefulWidget {
  final String field;

  @override
  _SliderRFState createState() => _SliderRFState();

  SliderRF({this.field});
}

class _SliderRFState extends State<SliderRF> {
  var currentValue = [0.0];
  var _lowerValue, _upperValue;

  @override
  Widget build(BuildContext context) {
    var hasFocus = false;
    double width = 18;

    var model = PageRF.of(context)?.model;

    Libs.modelValidationFunctions[widget.field] = () {
      if (widget.field != null) {
        var value = Dynamic(model).get(widget.field);
        ValidationDataRF data = ValidationRF.validate(widget.field, model, value);
        setState(() {
          Libs.modelValidationData[widget.field] = data;
        });
        return data?.isValid;
      }
      return false;
    };

    var value = Dynamic(model).get(widget.field);
    if (value != null) currentValue = [value];

    var slider = Container(
      margin: const EdgeInsets.only(right: 15),
      //color: Colors.grey,
      constraints: BoxConstraints.expand(height: 40),
      child: Stack(
        children: <Widget>[
          Positioned.fill(
            top: width,
            bottom: width,
            child: Container(
              //color: Colors.yellow,
              height: 5, //width: MediaQuery.of(context).size.width,
              child: Row(
                mainAxisAlignment: MainAxisAlignment.spaceBetween,
                children: <Widget>[
                  Container(
                    width: 20,
                    height: 3.5,
                    color: Colors.blue,
                    child: Text(' '),
                  ),
                  Container(
                    width: 20,
                    height: 3.5,
                    color: Color(0xfeeeeff),
                    child: Text(' '),
                  )
                ],
              ),
            ),
          ),
          Positioned(
            child: FlutterSlider(
              trackBar: FlutterSliderTrackBar(),
              values: currentValue,
              max: 500,
              min: 0,
              onDragStarted: (handlerIndex, lowerValue, upperValue) {
                FocusScope.of(context).requestFocus(FocusNode());
                print('ds');
              },
              onDragging: (handlerIndex, lowerValue, upperValue) {
                _lowerValue = lowerValue;
                _upperValue = upperValue;

                setState(() {
                  currentValue[0] = lowerValue;
                  print(currentValue[0]);
                  Dynamic(model).set(widget.field, currentValue[0]);
                  if (Libs.modelValidationData[widget.field] != null)
                    Libs.modelValidationData[widget.field].isValid = true;
                });
              },
            ),
          ),
        ],
      ),
    );

    return Column(
      crossAxisAlignment: CrossAxisAlignment.start,
      children: <Widget>[slider, ValidationRF.showErrorMessage(widget.field)],
    );
  }
}

//endregion SliderRF

//region RatingRF
class RatingRF extends StatefulWidget {
  final String field;

  @override
  _RatingRFState createState() => _RatingRFState();

  RatingRF({this.field});
}

class _RatingRFState extends State<RatingRF> {
  double initialRating = 0;

  @override
  Widget build(BuildContext context) {
    var model = PageRF.of(context)?.model;

    Libs.modelValidationFunctions[widget.field] = () {
      if (widget.field != null) {
        var value = Dynamic(model).get(widget.field);
        ValidationDataRF data = ValidationRF.validate(widget.field, model, value);
        setState(() {
          Libs.modelValidationData[widget.field] = data;
        });
        return data?.isValid;
      }
      return false;
    };

    var value = Dynamic(model).get(widget.field);
    if (value != null) initialRating = value;

    var rating = Container(
      height: 40,
      child: Stack(
        overflow: Overflow.visible,
        children: <Widget>[
          Positioned.fill(
            top: 2,
            left: -10,
            child: RatingBar.builder(
              initialRating: initialRating,
              minRating: 1,
              direction: Axis.horizontal,
              allowHalfRating: true,
              itemCount: 5,
              itemSize: 30,
              itemPadding: EdgeInsets.symmetric(horizontal: 4.0),
              itemBuilder: (context, index) => Icon(
                Icons.star,
                color: Colors.amber,
              ),
              onRatingUpdate: (rating) {
                FocusScope.of(context).requestFocus(FocusNode());
                print(rating);
                Dynamic(model).set(widget.field, rating);
                setState(() {
                  if (Libs.modelValidationData[widget.field] != null)
                    Libs.modelValidationData[widget.field].isValid = true;
                });
              },
            ),
          ),
        ],
      ),
    );

    return Column(
      crossAxisAlignment: CrossAxisAlignment.start,
      children: <Widget>[rating, ValidationRF.showErrorMessage(widget.field)],
    );
  }
}

//endregion RatingRF

//region SwitchRF
class SwitchRF extends StatefulWidget {
  final String field;
  final String label;

  @override
  _SwitchRFState createState() => _SwitchRFState();

  SwitchRF({this.label = "Active", this.field});
}

class _SwitchRFState extends State<SwitchRF> {
  int isSwitched = 0;

  @override
  Widget build(BuildContext context) {
    var model = PageRF.of(context)?.model;

    Libs.modelValidationFunctions[widget.field] = () {
      if (widget.field != null) {
        var value = Dynamic(model).get(widget.field);
        ValidationDataRF data = ValidationRF.validate(widget.field, model, value);
        setState(() {
          Libs.modelValidationData[widget.field] = data;
        });
        return data?.isValid;
      }
      return false;
    };

    var value = Dynamic(model).get(widget.field);
    if (value != null) isSwitched = value;

    var toggleSwitch = Row(
      children: <Widget>[
        Text(widget.label),
        SizedBox(
          width: widget.label == '' ? 0 : 28,
        ),
        Switch(
          value: isSwitched == 0 ? false : true,
          onChanged: (value) {
            setState(() {
              FocusScope.of(context).requestFocus(FocusNode());
              isSwitched = value ? 1 : 0;
              Dynamic(model).set(widget.field, isSwitched);
              if (Libs.modelValidationData[widget.field] != null) Libs.modelValidationData[widget.field].isValid = true;
            });
          },
//          activeTrackColor: Colors.lightGreenAccent,
//          activeColor: Colors.green,
        ),
      ],
    );

    return Column(
      crossAxisAlignment: CrossAxisAlignment.start,
      children: <Widget>[toggleSwitch, ValidationRF.showErrorMessage(widget.field)],
    );
  }
}

//endregion SwitchRF

//region RadioGroup
class RadioCheckData {
  final String label;
  final dynamic value;

  RadioCheckData({this.label, this.value});
}

class RadioRF extends StatefulWidget {
  final String field;
  final int value;
  final List<RadioCheckData> data;
  final dynamic dimension;
  final bool scrollable;
  final double height;
  final Function onChange;
  final bool separator;
  final double paddingTop;

  @override
  _RadioRFState createState() => _RadioRFState();

  RadioRF({
    this.data,
    this.dimension,
    this.field,
    this.scrollable,
    this.height,
    this.value,
    this.onChange,
    this.separator = false,
    this.paddingTop = 0,
  });
}

class _RadioRFState extends State<RadioRF> {
  int selectedValue = -1; // value could be String or anything ?
  double baseHeight = 40;

  @override
  void didChangeDependencies() {
    var model = PageRF.of(context)?.model;

    var value = (widget.field == null)
        ? widget.value ?? null
        : (model == null)
            ? null
            : Dynamic(model).get(widget.field);

    //var value = Dynamic(model).get(widget.field);
    if (value != null) selectedValue = value.runtimeType == int ? value : int.parse(value);
    if (widget.onChange != null) widget.onChange(value);
    super.didChangeDependencies();
  }

  @override
  Widget build(BuildContext context) {
    var model = PageRF.of(context)?.model;

    Libs.modelValidationFunctions[widget.field] = () {
      if (widget.field != null) {
        var value = Dynamic(model).get(widget.field);
        ValidationDataRF data = ValidationRF.validate(widget.field, model, value);
        setState(() {
          Libs.modelValidationData[widget.field] = data;
        });
        return data?.isValid;
      }
      return false;
    };

    updateSelectionStatus(int val) {
      setState(() {
        FocusScope.of(context).requestFocus(FocusNode());
        selectedValue = val;
        if (model != null) Dynamic(model).set(widget.field, selectedValue);
        if (widget.onChange != null) widget.onChange(selectedValue.toString());
        if (Libs.modelValidationData[widget.field] != null) Libs.modelValidationData[widget.field].isValid = true;
      });
    }

    var radioList = Stack(
      overflow: Overflow.visible,
      children: <Widget>[],
    );
    var _baseHeight = generateRadioList(widget.data, selectedValue, updateSelectionStatus, radioList);

    setState(() {
      baseHeight = _baseHeight;
    });

    return Column(
      crossAxisAlignment: CrossAxisAlignment.start,
      children: <Widget>[
        Container(constraints: BoxConstraints.expand(height: baseHeight), child: radioList),
        ValidationRF.showErrorMessage(widget.field)
        //Text('test')
      ],
    );
  }

  double generateRadioList(dynamic _data, dynamic _selectedValue, _fn, dynamic _radioStack) {
    double _height = 0;
    double offsetTop;
    double offsetTopHeight = 40;

    _radioStack.children.clear();
    for (RadioCheckData item in _data) {
      if (offsetTop == null)
        offsetTop = -10;
      else
        offsetTop = offsetTop + offsetTopHeight + (widget.separator ? (item == _data[_data.length - 1] ? 8 : 16) : 2);

      var pos = Positioned(
          top: offsetTop,
          left: -11,
          child: Column(
            crossAxisAlignment: CrossAxisAlignment.start,
            children: <Widget>[
              SizedBox(height: widget.paddingTop),
              Row(
                children: <Widget>[
                  Radio(
                      materialTapTargetSize: MaterialTapTargetSize.shrinkWrap,
                      value: item.value,
                      groupValue: _selectedValue,
                      onChanged: (value) {
                        _fn(value);
                        print(value);
                      }),
                  //Text(item.label, style: TextStyle(fontWeight: FontWeight.bold))
                  CheckRadioLabelRF(
                    index: item.value,
                    label: item.label,
                    onChange: (value) {
                      _fn(int.parse(value));
                      print(value.toString());
                    },
                  )
                ],
              ),
              Container(height: 5),
              widget.separator
                  ? item == _data[_data.length - 1]
                      ? Container()
                      : Container(
                          height: 1,
                          color: Colors.black12,
                          width: MediaQuery.of(context).size.width - 100,
                        )
                  : Container(),
            ],
          ));

      _height = _height + offsetTopHeight;
      _radioStack.children.add(pos);
    }
    return _height + 10; // 10 for click last item
  }
}
//endregion RadioGroup

//region CheckboxGroup
class CheckRF extends StatefulWidget {
  final List<RadioCheckData> data;
  final dynamic dimension;
  final String field;
  final String value;
  final bool scrollable;
  final double height;
  final Function onChange;
  final bool separator;
  final double paddingTop;

  @override
  _CheckRFState createState() => _CheckRFState();

  CheckRF({
    this.data,
    this.dimension,
    this.field,
    this.value,
    this.scrollable,
    this.height,
    this.onChange,
    this.separator = false,
    this.paddingTop = 0,
  });
}

class _CheckRFState extends State<CheckRF> {
  Map<int, bool> currentValueMap = Map();
  double baseHeight = 40;

  @override
  void didChangeDependencies() {
    var model = PageRF.of(context)?.model;

    // set value
    var value = (widget.field == null)
        ? widget.value ?? null
        : (model == null)
            ? null
            : Dynamic(model).get(widget.field);

    if (value != null) {
      Map<int, bool> currentValueMapFromString = Map();
      for (var item in value.toString().split(RegExp(','))) {
        if (!item.isEmpty) currentValueMapFromString[int.parse(item)] = true;
      }
      currentValueMap = currentValueMapFromString;
      if (widget.onChange != null) widget.onChange(value);
    }
    super.didChangeDependencies();
  }

  @override
  Widget build(BuildContext context) {
    var model = PageRF.of(context)?.model;

    Libs.modelValidationFunctions[widget.field] = () {
      if (widget.field != null) {
        var value = Dynamic(model).get(widget.field);
        ValidationDataRF data = ValidationRF.validate(widget.field, model, value);
        setState(() {
          Libs.modelValidationData[widget.field] = data;
        });
        return data?.isValid;
      }
      return false;
    };

    // on Select
    updateSelectionStatus(int value, bool val) {
      var string = '';
      setState(() {
        FocusScope.of(context).requestFocus(FocusNode());
        currentValueMap[value] = val;
        currentValueMap.forEach((k, v) {
          if (v) string += '${k.toString()},';
        });
        print(string);
        if (model != null) if (widget.field != null) Dynamic(model).set(widget.field, string);
        if (widget.onChange != null) widget.onChange(string);
        if (Libs.modelValidationData[widget.field] != null) Libs.modelValidationData[widget.field].isValid = true;
      });
    }

    var checkStack = Stack(
      overflow: Overflow.visible,
      children: <Widget>[],
    );
    var _baseHeight = generateCheckList(widget.data, currentValueMap, updateSelectionStatus, checkStack);

    setState(() {
      baseHeight = _baseHeight;
    });

    return Column(
      crossAxisAlignment: CrossAxisAlignment.start,
      children: <Widget>[
        Container(
          constraints: BoxConstraints.expand(height: baseHeight),
          child: checkStack,
          color: Colors.transparent,
        ),
        ValidationRF.showErrorMessage(widget.field)
      ],
    );
  }

  double generateCheckList(List<RadioCheckData> _data, Map<int, bool> _currentValueMap, _fn, dynamic _checkStack) {
    double _baseHeight = 0;
    double offsetTop;
    double offsetTopHeight = 18;
    double spaceBetweenRow = 20;

    _checkStack.children.clear();
    for (RadioCheckData item in _data) {
      if (offsetTop == null) {
        offsetTop = -15;
      } else {
        offsetTop = offsetTop + offsetTopHeight + (!widget.separator ? spaceBetweenRow : 2);
      }

      var pos = Positioned(
          top: offsetTop,
          left: -14,
          child: Column(
            crossAxisAlignment: CrossAxisAlignment.start,
            children: <Widget>[
              SizedBox(height: widget.paddingTop),
              Row(
                children: <Widget>[
                  Checkbox(
                      value: _currentValueMap[item.value] ?? false, onChanged: (newValue) => _fn(item.value, newValue)),
                  //Text(item.label, style: TextStyle(fontWeight: FontWeight.bold))
                  CheckRadioLabelRF(
                    index: item.value,
                    label: item.label,
                    onChange: (value) {
                      _fn(value, _currentValueMap[value] == null ? false : !_currentValueMap[value]);
                    },
                  )
                ],
              ),
              Container(
                height: 15,
                color: Colors.green,
              ),
              SizedBox(height: 100),
              (widget.separator)
                  ? (item == _data[_data.length - 1])
                      ? Container()
                      : Container(
                          height: 1,
                          color: Colors.black12,
                          width: MediaQuery.of(context).size.width - 100,
                        )
                  : Container(),
            ],
          ));

      _baseHeight = _baseHeight + offsetTopHeight;
      _checkStack.children.add(pos);
    }
    return _baseHeight + spaceBetweenRow;
  }
}

class CheckRadioLabelRF extends StatefulWidget {
  int index;
  String label;
  TextStyle style;
  Function onChange;

  CheckRadioLabelRF({this.index, this.label, this.style, this.onChange});

  @override
  _CheckRadioLabelRFState createState() => _CheckRadioLabelRFState();
}

class _CheckRadioLabelRFState extends State<CheckRadioLabelRF> {
  @override
  Widget build(BuildContext context) {
    return Material(
      color: Colors.transparent,
      child: InkWell(
        onTap: () {
          widget.onChange(widget.index.toString());
        },
        child: Container(
          width: 280,
          child: Padding(
            padding: const EdgeInsets.only(top: 9.0, bottom: 8.0),
            child: TextRF(widget.label, style: widget.style ?? TextStyle(fontWeight: FontWeight.bold)),
          ),
        ),
      ),
    );
  }
}

//CheckboxGroup(
//orientation: GroupedButtonsOrientation.VERTICAL,
//padding: EdgeInsets.only(left: 0),
//margin: EdgeInsets.all(0),
//labels: <String>[
//"Sports",
//"Reading",
//],
//onSelected: (List<String> checked) => print(checked.toString()),
//
//)
//endregion CheckboxGroup

//region GridViewRF
const gridPadding = EdgeInsets.all(10);

class GridViewRF extends StatefulWidget {
  final dynamic itemList;
  final dynamic child;
  final EdgeInsetsGeometry padding;
  final dynamic decoration;
  final int column;
  final double spaceX;
  final double spaceY;
  final bool isScroll;
  final GridBuild build;

  const GridViewRF(
      {this.itemList,
      this.child,
      this.padding = gridPadding,
      this.decoration,
      this.column = 2,
      this.spaceX = 2.0,
      this.spaceY = 2.0,
      this.isScroll = true,
      this.build});

  @override
  _GridViewRF createState() => _GridViewRF();
}

typedef GridBuild = Widget Function(BuildContext context, int index);

class _GridViewRF extends State<GridViewRF> {
  @override
  Widget build(BuildContext context) {
    dynamic modelList = widget.itemList;

    return Container(
        padding: widget.padding,
        child: GridView.builder(
          physics: !widget.isScroll ? const NeverScrollableScrollPhysics() : null,
          itemCount: modelList.length,
          gridDelegate: SliverGridDelegateWithFixedCrossAxisCount(
              crossAxisCount: widget.column, crossAxisSpacing: widget.spaceY, mainAxisSpacing: widget.spaceX),
          itemBuilder: (BuildContext context, int index) {
            return widget.build(context, index);
          },
        ));
  }
}
//endregion

//#region FilterRF

class FilterRF extends StatefulWidget {
  Function onAdvanceFilter;
  DataSourceRF dataSource;
  Color color;
  String quickFilterFields;
  String defaultSorting;
  bool showAdvanceButton;
  Function onCustomActionTap;

  @override
  _FilterRFState createState() => _FilterRFState();

  FilterRF({
    this.dataSource,
    this.onAdvanceFilter,
    this.quickFilterFields,
    this.defaultSorting = 'DESC',
    this.color,
    this.showAdvanceButton = true,
    this.onCustomActionTap,
  });
}

class _FilterRFState extends State<FilterRF> {
  var quickFilterValue;
  final textController = TextEditingController();

  @override
  Widget build(BuildContext context) {
    return Container(
      margin: const EdgeInsets.only(left: 20, right: 20),
      child: Row(
        children: [
          Expanded(
            child: Container(
              height: 42,
              decoration: BoxDecoration(
                color: widget.color ?? Colors.grey[200],
                borderRadius: BorderRadius.all(Radius.circular(8)),
              ),
              child: Row(
                children: [
                  Padding(
                    padding: const EdgeInsets.all(8.0),
                    child: Icon(Icons.search, color: Colors.grey),
                  ),
                  Expanded(
                    child: Center(
                      child: TextField(
                        autofocus: false,
                        controller: textController,
                        decoration: InputDecoration(border: InputBorder.none, hintText: 'Search'),
                        onChanged: (input) async {
                          quickFilterValue = input;
                          widget.dataSource.currentPage = 1;
                          await widget.dataSource
                              .filterQuick('${widget.quickFilterFields}#$quickFilterValue#${widget.defaultSorting}');
                          //setState(() {});
                        },
                      ),
                    ),
                  ),
                ],
              ),
            ),
          ),
          SizedBox(
            width: 10,
          ),
          widget.showAdvanceButton
              ? InkWell(
                  onTap: () {
                    FocusScope.of(context).requestFocus(FocusNode());
                    widget.dataSource.currentPage = 1;
                    widget.onAdvanceFilter();
                  },
                  child: Container(
                    //advance filter
                    width: 40,
                    height: 40,
                    child: Padding(
                      padding: const EdgeInsets.all(8.0),
                      child: Image.asset("assets/images/settings.png", color: Colors.grey),
                    ),
                    decoration: BoxDecoration(
                        color: widget.color ?? Colors.white, //shape: BoxShape.circle,
                        boxShadow: [
                          BoxShadow(
                            color: Colors.grey.withOpacity(0.3),
                            blurRadius: 3.0,
                            spreadRadius: 1.0,
                          ),
                        ],
                        borderRadius: BorderRadius.all(Radius.circular(10))),
                  ),
                )
              : Container(),
          SizedBox(
            width: 10,
          ),
          InkWell(
            onTap: () {
              FocusScope.of(context).requestFocus(FocusNode());
              widget.dataSource.filterClear();
              textController.text = '';
            },
            child: Container(
              width: 40,
              height: 40,
              child: Icon(Icons.refresh, color: Colors.grey),
              decoration: BoxDecoration(
                  color: widget.color ?? Colors.white, //shape: BoxShape.circle,
                  boxShadow: [
                    BoxShadow(
                      color: Colors.grey.withOpacity(0.3),
                      blurRadius: 3.0,
                      spreadRadius: 1.0,
                    ),
                  ],
                  borderRadius: BorderRadius.all(Radius.circular(10))),
            ),
          ),
          SizedBox(width: 10),
          InkWell(
            onTap: () {
              FocusScope.of(context).requestFocus(FocusNode());
              widget.onCustomActionTap("geo");
            },
            child: Container(
              width: 40,
              height: 40,
              child: Icon(Icons.my_location, color: Colors.grey),
              decoration: BoxDecoration(
                  color: widget.color ?? Colors.white, //shape: BoxShape.circle,
                  boxShadow: [
                    BoxShadow(
                      color: Colors.grey.withOpacity(0.3),
                      blurRadius: 3.0,
                      spreadRadius: 1.0,
                    ),
                  ],
                  borderRadius: BorderRadius.all(Radius.circular(10))),
            ),
          )
        ],
      ),
    );
  }
}
//#endregion FilterRF

//#region ListViewRFItems
class ListCardRF extends StatefulWidget {
  final leading;
  final body;
  final action;
  final statusColor;
  final bottomLevel;
  final EdgeInsets padding;

  @override
  _ListCardRFState createState() => _ListCardRFState();

  ListCardRF({
    this.leading,
    this.statusColor = Colors.green,
    this.body,
    this.action,
    this.bottomLevel,
    this.padding,
  });
}

class _ListCardRFState extends State<ListCardRF> {
  @override
  Widget build(BuildContext context) {
    return Container(
      margin: EdgeInsets.only(
        top: widget?.padding?.top ?? 5,
        bottom: widget?.padding?.bottom ?? 10,
        left: widget?.padding?.left ?? 10,
        right: widget?.padding?.right ?? 10,
      ),
      child: Container(
        padding: EdgeInsets.only(top: 8),
        child: Container(
          decoration: BoxDecoration(
              boxShadow: [
                BoxShadow(
                  color: Colors.black.withOpacity(0.1),
                  spreadRadius: 0,
                  blurRadius: 5,
                  offset: Offset(0, -5), // changes position of shadow
                ),
                BoxShadow(
                  color: Colors.black.withOpacity(0.2),
                  spreadRadius: 0,
                  blurRadius: 5,
                  offset: Offset(0, 5), // changes position of shadow
                ),
              ],
              color: widget.statusColor,
              borderRadius: new BorderRadius.only(
                  topLeft: Radius.circular(20),
                  bottomLeft: Radius.circular(20),
                  topRight: Radius.circular(20),
                  bottomRight: Radius.circular(20))),
          child: Container(
            height: 90,
            margin: EdgeInsets.only(left: 6),
            decoration: BoxDecoration(
                color: Colors.white,
                borderRadius: new BorderRadius.only(
                    topLeft: Radius.circular(20),
                    bottomLeft: Radius.circular(20),
                    topRight: Radius.circular(20),
                    bottomRight: Radius.circular(20))),
            child: Column(
              mainAxisAlignment: MainAxisAlignment.spaceEvenly,
              children: <Widget>[
                Expanded(
                  child: Stack(
                    children: [
                      Row(
                        crossAxisAlignment: CrossAxisAlignment.center,
                        children: <Widget>[
                          SizedBox(
                            width: 10,
                          ),
                          widget.leading,
                          Expanded(
                            child: Padding(
                              padding: EdgeInsets.only(left: 18.0, top: 5, right: 20),
                              child: widget.body,
                            ),
                            flex: 70,
                          ),
                        ],
                      ),
                      Positioned(
                        right: 0,
                        bottom: 0,
                        child: Column(
                          crossAxisAlignment: CrossAxisAlignment.end,
                          mainAxisAlignment: MainAxisAlignment.end,
                          children: [
                            Container(
                              margin: EdgeInsets.only(right: 12),
                              child: widget.action,
                            ),
                            Align(
                              alignment: Alignment.bottomRight,
                              child: Container(
                                height: 30,
                                width: 140,
                                decoration: BoxDecoration(
                                    color: Color(0Xffffb1b3),
                                    borderRadius: new BorderRadius.only(
                                      topLeft: Radius.circular(30),
                                      bottomRight: Radius.circular(20),
                                    )),
                                child: Padding(
                                  padding: const EdgeInsets.only(left: 12.0),
                                  child: Center(
                                    child: TextRF(
                                      widget.bottomLevel,
                                      size: 14,
                                      color: '7f3133'.toColor,
                                      //Colors.brown[800],
                                      weight: FontWeight.w600,
                                    ),
                                  ),
                                ),
                              ),
                            )
                          ],
                        ),
                      )
                    ],
                  ),
                ),
              ],
            ),
          ),
        ),
      ),
    );
  }
}

class CardAvatarRF extends StatefulWidget {
  @required
  final image;
  final borderColor;
  final borderWidth;
  final statusColor;

  @override
  _CardAvatarRFState createState() => _CardAvatarRFState();

  CardAvatarRF({
    this.image,
    this.borderColor,
    this.borderWidth,
    this.statusColor = Colors.green,
  });
}

class _CardAvatarRFState extends State<CardAvatarRF> {
  @override
  Widget build(BuildContext context) {
    return Stack(
      children: <Widget>[
        Container(
            width: 50.0,
            height: 50.0,
            decoration: new BoxDecoration(
              color: Color(0XffFC6767).withOpacity(0.3),
/*            image: new DecorationImage(
              image: AssetImage(widget.image),
              fit: BoxFit.cover,
            ),*/
              borderRadius: new BorderRadius.all(new Radius.circular(50.0)),
              /*    border: new Border.all(
              color: Colors.red,
              width: 2.0,
            ),*/
            ),
            child: Padding(
              padding: const EdgeInsets.all(8.0),
              child: Image(
                image: AssetImage(widget.image),
              ),
            )),
/*        Positioned(
          bottom: 0,
          right: 0,
          child: CircleAvatar(
            backgroundColor:  Colors.white,
            radius: 9,
            child: CircleAvatar(
              backgroundColor: widget.statusColor,
              radius: 7,
            ),
          ),
        )*/
      ],
    );
  }
}

class CardActionRF extends StatefulWidget {
  final icon;
  final deleteId;
  final index;
  final size;
  final color;
  final onAction;
  DataSourceRF dataSource;
  List<Function> modelValidators;

  @override
  _CardActionRFState createState() => _CardActionRFState();

  CardActionRF(this.icon, this.deleteId, this.dataSource, this.index,
      {this.color, this.size, this.onAction, this.modelValidators});
}

class _CardActionRFState extends State<CardActionRF> {
  @override
  Widget build(BuildContext context) {
    var model = Models.models[Dynamic.genericModelName(widget.dataSource.modelList)].model;
    return IconButton(
        icon: Icon(widget.icon),
        iconSize: 40,
        //color: Color.fromRGBO(140, 140, 140, 1),
        color: Colors.grey.shade400,
        onPressed: () {
          AlertRF(
              context: context,
              title: 'DeleteAlertTitle',
              icon: Icons.error,
              msg: 'DeleteALertDescription',
              confirmText: 'DELETE',
              onConfirm: (value) async {
                if (value) {
                  var ret = await rf.db.delete(model, widget.deleteId, widget.modelValidators);
                  if (ret == 0) {
                    rf.msg(context, 'DeleteALertSyncedData');
                  } else {
                    widget.dataSource.delete(widget.index);
                    widget.dataSource.refresh();
                    setState(() {});
                  }
                }
              }).show();
        }
        //widget.onAction,
        );
  }
}
//#endregion ListViewRFItems

//#region GridCardItems
class GridCardRF extends StatefulWidget {
  int index;
  List<String> cardTitleList;
  List<IconData> cardImageList;
  Function onTap;

  GridCardRF({
    this.index,
    this.cardTitleList,
    this.cardImageList,
    this.onTap,
  });

  @override
  _GridCardRFState createState() => _GridCardRFState();
}

class _GridCardRFState extends State<GridCardRF> {
  @override
  Widget build(BuildContext context) {
    return FlatButton(
      onPressed: () => widget.onTap(widget.index),
      color: Colors.white,
      child: Container(
        //color: Colors.red,
        child: Column(
          crossAxisAlignment: CrossAxisAlignment.center,
          mainAxisAlignment: MainAxisAlignment.center,
          children: <Widget>[
            FaIcon(widget.cardImageList[widget.index], size: 40, color: Colors.pink),
            SizedBox(height: 5),
            Center(
              child: Text(
                widget.cardTitleList[widget.index].toUpperCase(),
                textAlign: TextAlign.center,
                style: TextStyle(fontWeight: FontWeight.bold, fontSize: 15),
              ),
            )
          ],
        ),
      ),
    );
  }
}
//#endregion GridCardItems

//#region LabelAndTitles
class LabelRF extends StatefulWidget {
  final text;

  @override
  _LabelRFState createState() => _LabelRFState();

  LabelRF(this.text);
}

class _LabelRFState extends State<LabelRF> {
  @override
  Widget build(BuildContext context) {
    return TextRF(
      widget.text,
      style: TextStyle(fontWeight: FontWeight.bold, fontSize: 14),
    );
  }
}

class PageTitleRF extends StatefulWidget {
  final text;

  @override
  _PageTitleRFState createState() => _PageTitleRFState();

  PageTitleRF(this.text);
}

class _PageTitleRFState extends State<PageTitleRF> {
  @override
  Widget build(BuildContext context) {
    return Container(
      height: 40,
      child: Stack(
        overflow: Overflow.visible,
        children: <Widget>[
          Positioned(
            left: -10,
            child: Container(
                child: Padding(
                  padding: const EdgeInsets.only(left: 3, top: 5, bottom: 5),
                  child: Text(
                    widget.text,
                    style: TextStyle(
                      color: Colors.pink,
                      fontSize: 20,
                      fontWeight: FontWeight.bold,
                      fontFamily: 'Roboto',
                      letterSpacing: 1,
                      decoration: TextDecoration.none,
                    ),
                  ),
                ),
                decoration: BoxDecoration(
                    border: Border(
                        left: BorderSide(
                  color: Colors.pink,
                  width: 8,
                )))),
          )
        ],
      ),
    );
  }
}

class PageSubTitleRF extends StatefulWidget {
  final text;

  @override
  _PageSubTitleRFState createState() => _PageSubTitleRFState();

  PageSubTitleRF(this.text);
}

class _PageSubTitleRFState extends State<PageSubTitleRF> {
  @override
  Widget build(BuildContext context) {
    return Padding(
      padding: const EdgeInsets.only(bottom: 15.0),
      child: Text(
        widget.text,
        style: TextStyle(
          color: Colors.blueAccent,
          fontSize: 17,
          fontWeight: FontWeight.bold,
          fontFamily: 'Roboto',
          letterSpacing: 1,
          decoration: TextDecoration.none,
        ),
      ),
    );
  }
}
//#endregion LabelAndTitles

//#region GapRF
enum GapTypeRF { Gap5, Gap10 }

class GapRF extends StatefulWidget {
  double gap;

  GapRF([this.gap = 5]);

  @override
  _GapRFState createState() => _GapRFState();
}

class _GapRFState extends State<GapRF> {
  @override
  Widget build(BuildContext context) {
    return Container(
      height: widget.gap,
      width: widget.gap,
    );
  }
}
//#endregion GapRF

//#region GraphRF

enum GraphTypeRF { line, polar }

class GraphRF extends StatefulWidget {
  String modelName;
  String fieldName;
  GraphTypeRF graphType;

  GraphRF({this.modelName, this.fieldName, this.graphType = GraphTypeRF.line});

  @override
  _GraphRFState createState() => _GraphRFState();
}

class _GraphRFState extends State<GraphRF> {
  DataSourceRF dataSource;
  List<Map<String, dynamic>> graphData = [];

  Future<List<Map<String, dynamic>>> getData() async {
    Models models = Models();
    for (var element in models.get().entries) {
      if (element.key.toString() == widget.modelName) {
        String sql =
            "select COUNT(*) as sold,${widget.fieldName} as genre from ${widget.modelName} group by ${widget.fieldName}";
        print(sql);
        await rf.db.get(GraphItem, sql, null, (modellist) {
          for (int i = 0; i < modellist.length; i++) {
            graphData.add((modellist[i] as GraphItem).toJson());
          }
          print(graphData.toString());
        });
      }
    }
    return graphData;
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      height: 200,
      width: ScreenRF.width(context) - 20,
      child: FutureBuilder<List<Map<String, dynamic>>>(
        future: getData(),
        builder: (context, AsyncSnapshot<List<Map<String, dynamic>>> snapshot) {
          if (!snapshot.hasData) {
            return Container();
          } else {
            if (widget.graphType == GraphTypeRF.line) {
              return graphic.Chart(
                data: snapshot.data,
                scales: {
                  'genre': graphic.CatScale(
                    accessor: (map) => map['genre'] as String,
                  ),
                  'sold': graphic.LinearScale(
                    accessor: (map) => map['sold'] as num,
                    nice: true,
                  )
                },
                geoms: [
                  graphic.IntervalGeom(
                    position: graphic.PositionAttr(field: 'genre*sold'),
                    shape: graphic.ShapeAttr(values: [graphic.RectShape(radius: Radius.circular(5))]),
                  )
                ],
                axes: {
                  'genre': graphic.Defaults.horizontalAxis,
                  'sold': graphic.Defaults.verticalAxis,
                },
              );
            } else {
              return graphic.Chart(
                data: snapshot.data,
                scales: {
                  'genre': graphic.CatScale(
                    accessor: (map) => map['genre'] as String,
                  ),
                  'sold': graphic.LinearScale(
                    accessor: (map) => map['sold'] as num,
                    nice: true,
                  )
                },
                coord: graphic.PolarCoord(),
                geoms: [
                  graphic.IntervalGeom(
                    position: graphic.PositionAttr(field: 'genre*sold'),
                    color: graphic.ColorAttr(field: 'genre'),
                  )
                ],
                axes: {
                  'genre': graphic.Defaults.circularAxis,
                  'sold': graphic.Defaults.radialAxis..label = null,
                },
              );
            }
          }
        },
      ),
    );
    /*  locations.add(GraphItem(genre: "afdas", sold: 11).toJson());
    return Container(
      height: 200,
      width: ScreenRF.width(context) - 20,
      child: graphic.Chart(
        data: locations,
        scales: {
          'genre': graphic.CatScale(
            accessor: (map) => map['genre'] as String,
          ),
          'sold': graphic.LinearScale(
            accessor: (map) => map['sold'] as num,
            nice: true,
          )
        },
        geoms: [
          graphic.IntervalGeom(
            position: graphic.PositionAttr(field: 'genre*sold'),
          )
        ],
        axes: {
          'genre': graphic.Defaults.horizontalAxis,
          'sold': graphic.Defaults.verticalAxis,
        },
      ),
    );*/
  }
}

@annotation.Model
class GraphItem {
  String genre;
  int sold;
  GraphItem({this.genre, this.sold});

  Map<String, dynamic> toJson() {
    final Map<String, dynamic> data = new Map<String, dynamic>();
    data['genre'] = this.genre;
    data['sold'] = this.sold;
    return data;
  }
}

//#endregion GapRF

//#region PageRFItems
Widget fab(Function onPressed) {
  return FloatingActionButton(
    backgroundColor: Colors.red,
    elevation: 10,
    onPressed: () {
      onPressed();
      //navigateToDetail(Profile(), 'Add Profile');
    },
    tooltip: 'Add',
    child: Icon(
      Icons.add,
      color: Colors.white,
    ),
  );
}

class BounceScrollBehavior extends ScrollBehavior {
  @override
  getScrollPhysics(_) => const BouncingScrollPhysics();
}
//#endregion PageRFItems

//#region OnboardItems
enum RippleEffectRF { style1 }

class SplashRF {
  String title;
  String subTitle;
  Color color;
  String font;
  Color titleColor;
  Color subTitleColor;
  Color rippleColor;
  String rippleImagePath;
  double rippleImageSize;
  String orgLogoPath;
  int duration;

  SplashRF({
    this.title,
    this.subTitle,
    this.color,
    this.font,
    this.titleColor,
    this.subTitleColor,
    this.rippleColor,
    this.rippleImagePath,
    this.rippleImageSize,
    this.orgLogoPath,
    this.duration = 5,
  });
}

class IntroRF {
  Widget title;
  Widget image;
  Widget description;
  Widget pointerImage;
  Color backgroundColor;
  RippleEffectRF rippleEffect;
  Color rippleColor;

  IntroRF({
    this.title,
    this.image,
    this.description,
    this.pointerImage,
    this.backgroundColor,
    this.rippleEffect,
    this.rippleColor,
  });
}

class WidgetsRF {
  static selectWidget(local, clint, store) {
    var widget = Object();
    if (local is EmptyRF) {
      widget = null;
    } else if (local.runtimeType.toString().lastIndexOf('RF') > 1) {
      if (local is TextRF || local is ContainerRF || local is DrawerRF) {
        if (local?.template?.version == 'EmptyRF') {
          widget = null;
        }
      }
    }

    if (clint is EmptyRF) {
      widget = null;
    } else if (clint.runtimeType.toString().lastIndexOf('RF') > 1) {
      if (clint is TextRF || clint is ContainerRF || clint is DrawerRF) {
        if (clint?.template?.version == 'EmptyRF') {
          widget = null;
        }
      }
    }

    if (widget == null) {
      return null;
    } else {
      return local ?? clint ?? store;
    }
  }
}
//#endregion OnboardItems

//#region Experiment
class StoryboardRF {
  String title; //
  OutlineRF outline; //
  FlexBarRF flex; //
  //List<Widget> actions; ////
  Object model;
  List<DataSourceRF> dataSources;
  Widget stickyHeader;
  Widget child;
  final Widget fab; //
  List<DrawerItemRF> drawerItems;
  DrawerItemRF drawerFooter;
  List<Color> osBarColors;

  StoryboardRF(
      {this.title,
      this.outline,
      this.flex,
      this.stickyHeader,
      this.model,
      this.dataSources, //this.actions,
      this.child,
      this.fab,
      this.drawerItems,
      this.drawerFooter,
      this.osBarColors});
}

class PageBuilderRF extends StatefulWidget {
  final Function(BuildContext) builder;

  const PageBuilderRF({Key key, this.builder}) : super(key: key);

  @override
  PageBuilderRFState createState() => new PageBuilderRFState();

  static PageBuilderRFState of(BuildContext context) {
    return context.ancestorStateOfType(const TypeMatcher<PageBuilderRFState>());
  }
}

class PageBuilderRFState extends State<PageBuilderRF> {
  @override
  Widget build(BuildContext context) {
    return widget.builder(context);
  }

  void rebuild() {
    setState(() {});
  }
}
//#endregion Experiment

//region BackupOld
/*
Stack(
        children: [
        Container(
          width: 10,
            height: 90,
            decoration:BoxDecoration(color: Colors.red, borderRadius: new BorderRadius.only(topLeft:Radius.circular(10),bottomLeft: Radius.circular(10))),
          ),
        Container(
          height: 90,
          margin: EdgeInsets.only(left:10),
          decoration:BoxDecoration(color:  Colors.white, borderRadius: new BorderRadius.only(topRight:Radius.circular(10),bottomRight:Radius.circular(10))),
          child: Column(
            mainAxisAlignment:MainAxisAlignment.spaceEvenly ,
            children: <Widget>[
              Expanded(
                child: Row(
                  crossAxisAlignment: CrossAxisAlignment.center,
                  children: <Widget>[
                    SizedBox(width: 10,),
                    widget.leading,
                    Expanded(
                      child: Padding(
                        padding: EdgeInsets.only(left: 18.0, top: 5),
                        child: widget.body,
                      ),
                      flex: 70,
                    ),
                    Expanded(
                      child: Column(
                        crossAxisAlignment: CrossAxisAlignment.end,
                        mainAxisAlignment: MainAxisAlignment.end,
                        children: [
                        widget.action,
                        Align(
                          alignment: Alignment.bottomRight,
                          child: Container(
                            height: 30,
                            decoration:BoxDecoration(color: Color(0XffF6E3B1), borderRadius: new BorderRadius.only(topLeft:Radius.circular(30),bottomRight:Radius.circular(10))),
                            child: Padding(
                              padding: const EdgeInsets.only(left:12.0),
                              child: Center(child: TextRF("December 20.2020",color: Color(0XffFF8A00),weight: FontWeight.w300,size: 12,),),
                            ),
                          ),
                        )
                      ],),
                      flex: 40,
                    ),

                  ],
                ),
              ),
            ],
          ),
        ),
      ],

      )
*/
//endregion BackupOld
